import std;

#include <windows.h>

int WINAPI wWinMain([[maybe_unused]] HINSTANCE instance, HINSTANCE /*prev_instance*/,
					LPWSTR /*cmd_line*/, int show_command)
{
	//LPCWSTR is wchar_t const *
	LPCWSTR class_name = L"My Window Class";

	//Description of our custom window class, for details see:
	//https://learn.microsoft.com/en-us/windows/win32/api/winuser/ns-winuser-wndclassexw
	WNDCLASSEXW wcx {
		.cbSize = sizeof(WNDCLASSEXW), //May be used by windows for API versioning. If programs compiled against
									   //different versions of Windows API headers end up having different variants of
									   //this structure, the size field will allow Windows to adapt dynamically and
									   //interpret it correctly for all variants. This technique is used in
									   //many WinAPI structures.

		.lpfnWndProc = DefWindowProcW, //window procedure - function used to handle messages intended for our class of
									   //windows. The default one provided by WinAPI handles some basic interactions.

		.hInstance = instance,		   //We should specify our program's instance handle, since it is used to identify
									   //the class along with the class name, but for now our program would work without
									   //this line

		.lpszClassName = class_name	   //Name of our custom class (should be unique in our program).
	};

	//Creating the class, for details see:
	//https://learn.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-registerclassexw
	ATOM class_id = RegisterClassExW(&wcx);
	if (!class_id)
	{
		OutputDebugStringW(L"Failed to register window class!");
		return -1;
	}

	//Create a window with our custom class; returns a handle we can use to operate on our window, for details see:
	//https://learn.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-createwindowexw
	HWND window = CreateWindowExW(0 /* extended window styles controlling its look and behaviour */,
								  class_name /*or class_id - window class */,
								  L"Hello First Window!" /* window title shown in the caption bar */,
								  0, /* window styles controlling its look and behaviour */
								  100, 100, /* x and y of window's position in screen coordinates */
								  640, 360, /* width and height of the window including borders, caption bar, etc. */
								  nullptr /* parent window's handle */,
								  nullptr /* main menu handle */,
								  nullptr /* or instance - used to identify module (executable or dll) 
											 for which the window class is registered */,
								  nullptr /* custom pointer we can use for our own data - we'll show its use later */);
	if (!window)
	{
		OutputDebugStringW(L"Failed to create a window!");
		return -1;
	}

	//Show the window on the desktop
	ShowWindow(window, show_command);

	//Message Loop - processes messages directed to our program, e.g. to allow for user interactions with our window,
	//for details see: https://learn.microsoft.com/en-us/windows/win32/winmsg/about-messages-and-message-queues
	MSG msg{};
	BOOL result;
	while ((result = GetMessageW(&msg,
								 nullptr /* window handle [optional] - used to filter messages by recipient window */,
								 0, 0 /* min and max message type - used to filter messages by type,
										 0, 0 - means no filter */)) != 0)
	{
		if (result == -1)
		{
			OutputDebugStringW(L"Failed to retrieve a message!");
			return -1;
		}
		//Here we can do some general message processing, but usually they are forwarded to the window procedure
		//associated with the recipient window:
		DispatchMessageW(&msg);
	}
	//Window should be destroyed before program exits, however our program will never exit unless GetMessageW fails
	//which is highly unlikely - even if we somehow close the window, the message loop will continue
	std::unreachable();
	//return static_cast<int>(msg.wParam);
}