#include <windows.h>

//Default window procedure is not sufficient, we must define our own, for details see:
//https://learn.microsoft.com/en-us/windows/win32/winmsg/about-window-procedures
//https://learn.microsoft.com/en-us/windows/win32/api/winuser/nc-winuser-wndproc
LRESULT CALLBACK WindowProcedure(HWND window /* recipient window handle */,
								 UINT msg_type /* message type identifier */,
								 WPARAM w_param, LPARAM l_param /* message parameters,
																   interpretation depends on message type*/)
{
	switch (msg_type)
	{
	case WM_CLOSE:
		//window was closed, we should destroy it
		//https://learn.microsoft.com/en-us/windows/win32/winmsg/wm-close
		DestroyWindow(window);//this will indirectly call our window procedure again with WM_DESTROY message
		return 0;

	case WM_DESTROY:
		//window was destroyed, so we should exit our message loop
		//https://learn.microsoft.com/en-us/windows/win32/winmsg/wm-destroy
		PostQuitMessage(0); //this posts WM_QUIT to our message queue
		return 0;

	default:
		//forward all other messages to default window procedure to preserve the built-in interactions
		return DefWindowProcW(window, msg_type, w_param, l_param);
	}
}

int WINAPI wWinMain([[maybe_unused]] HINSTANCE instance, HINSTANCE,	LPWSTR, int /* show_command */)
{
	LPCWSTR class_name = L"My Window Class";

	WNDCLASSEXW wcx{
		.cbSize = sizeof(WNDCLASSEXW),
		.lpfnWndProc = WindowProcedure /* use the custom window procedure for windows of our class */,
		.hCursor = LoadCursorW(nullptr, IDC_ARROW) /* set the mouse cursor for window's client area to an arrow */,
		.hbrBackground = reinterpret_cast<HBRUSH>(COLOR_WINDOW + 1)/* Brush handle used to repaint window's client area;
																	  without it parts of the window will become black
																	  when resizing. Some default brushes can be
																	  obtained by adding 1 to one of standard system
																	  color constants */,
		.lpszClassName = class_name
	};

	if (!RegisterClassExW(&wcx))
	{
		OutputDebugStringW(L"Failed to register window class!");
		return -1;
	}

	//In most circumstances it is more convenient to specify the size of window's client area (area inside the window).
	//However, CreateWindowExW takes the size of the whole window (including border, caption bar, menu and scrollbars).
	//We can convert from one to the other using AdjustWindowRectEx, specifying which elements the window will have
	//(with exception of scrollbars, this function is unable to account for those). For details, see:
	//https://learn.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-adjustwindowrectex
	RECT size{ 0, 0, 640, 360 };
	AdjustWindowRectEx(&size, WS_OVERLAPPEDWINDOW, false, 0);

	//Missing window elements are enabled using window styles, for details see:
	//https://learn.microsoft.com/en-us/windows/win32/winmsg/window-styles and
	//https://learn.microsoft.com/en-us/windows/win32/winmsg/extended-window-styles
	HWND window = CreateWindowExW(0, class_name, L"Hello First Window (fixed :D)!",
								  WS_OVERLAPPEDWINDOW /* specify style for a top-level window which includes:
														 caption bar, resizing border,
														 system menu (also enables the close button),
														 minimize and maximize buttons */
								  | WS_VISIBLE, /* combine with visible to avoid having to call ShowWindow - 
												   we can't specify show_command now, but Windows will use that
												   value anyway */
								  100, 100, size.right - size.left, size.bottom - size.top,
								  nullptr, nullptr, nullptr, nullptr);
	if (!window)
	{
		OutputDebugStringW(L"Failed to create a window!");
		return -1;
	}

	MSG msg{};
	BOOL result;
	//When WM_QUIT is received, GetMessageW returns 0
	while (0 != (result = GetMessageW(&msg, nullptr, 0, 0)))
	{
		if (result == -1)
		{
			OutputDebugStringW(L"Failed to retrieve a message!");
			return -1;
		}
		DispatchMessageW(&msg);


		//Final Note!
		//Not all messages pass through our custom message loop. Some, like WM_DESTROY, do not pass through the queue
		//at all (see the window procedure above). There are also some modal interactions, like window resizing, or 
		//showing a message box, that process queued messages internally. Therefore custom message processing in the
		//message loop should be avoided if possible, and should instead happen in the window procedure, since all
		//messages for our window will generally end up there one way or another.
	}
	//for WM_QUIT message, wParam contains the exit code passed to PostQuitMessage
	return static_cast<int>(msg.wParam);
}