module;

#include <windows.h>

export module pige.winapi:window;

export namespace pige::winapi {

	class window
	{
	private:

		static void register_class(HINSTANCE instance, LPCWSTR class_name)
		{
			WNDCLASSEXW wcx{
				.cbSize = sizeof(WNDCLASSEXW),
				.style = CS_VREDRAW | CS_HREDRAW /* redraw the entire client area on resize */,
				.lpfnWndProc = window_proc /* use static window_proc member function as window procedure, it will
											  forward messages to instance member function */,
				.hInstance = instance,
				.hCursor = LoadCursorW(nullptr, IDC_ARROW),
				.hbrBackground = static_cast<HBRUSH>(GetStockObject(WHITE_BRUSH)) /* some predefined brushes can be
																					 obtained from GetStockObject, see:
								 https://learn.microsoft.com/en-us/windows/win32/api/wingdi/nf-wingdi-getstockobject */,
				.lpszClassName = class_name,
			};
			RegisterClassExW(&wcx);
		}

		[[nodiscard]] static bool is_class_registered(HINSTANCE instance, LPCWSTR class_name)
		{
			WNDCLASSEXW wcx{};
			//We can try to retrieve class description using GetClassInfoExW.
			//This will tell us if it already has been registered. For details, see:
			//https://learn.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-getclassinfoexw
			return GetClassInfoExW(instance, class_name, &wcx);
		}

		HWND m_handle;
	public:
		/***** Constructors *****/
		inline window() : m_handle{ nullptr } { }
		window(window const &) = delete;
		window(HINSTANCE instance, LPCWSTR title)
			: window{ }
		{

			LPCWSTR class_name = L"My Window Class";
			if (!is_class_registered(instance, class_name))
				register_class(instance, class_name);
			CreateWindowExW(0, class_name, title, WS_OVERLAPPEDWINDOW | WS_VISIBLE,
							//pass CW_USEDEFAULT for position and size to allow Windows to choose them
							CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
							nullptr, nullptr, nullptr,
							reinterpret_cast<LPVOID>(this) /*custom parameter passes this to WM_NCCREATE*/);

			//m_handle will be set in WM_NCCREATE
		}

		/***** Window Procedure *****/
		static LRESULT CALLBACK window_proc(HWND handle, UINT msg, WPARAM wparam, LPARAM lparam)
		{
			//Static window procedure will forward messages to instance window_proc, but we need a pointer to
			//the window instance associated with handle
			window *w = nullptr;
			if (msg == WM_NCCREATE)
			{
				//WM_NCCREATE is one of the first messages sent to the window during CreateWindowExW, and the first
				//where we have access to the parameters of that function. That is the earliest place where we have both
				//the window handle and the pointer to the window class instance, so we can associate them with one
				//another. CreateWindowExW parameters are stored in a CREATESTRUCTW, a pointer to which is in lparam.
				auto pcs = reinterpret_cast<LPCREATESTRUCTW>(lparam);
				w = reinterpret_cast<window *>(pcs->lpCreateParams /* this pointer from constructor */);
				//to get from window object to handle we will store it in the m_handle field
				w->m_handle = handle;
				//We also need a way to get from handle to window object.
				//The window data stored internally by Windows has a pointer-sized space we can use for any purpose.
				//We will use it to store the pointer to window object, that can later be retrieved at any time.
				SetWindowLongPtrW(handle, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(w));
			}
			else {
				//For any other message try to retrieve the object pointer from GWLP_USERDATA
				w = reinterpret_cast<window *>(GetWindowLongPtrW(handle, GWLP_USERDATA));
			}
			if (w)
			{
				auto r = w->window_proc(msg, wparam, lparam);
				if (msg == WM_NCDESTROY)
					w->m_handle = nullptr;
				//No need to set GLWP_USERDATA to 0, as internall window data no longer exists
				return r;
			}
			return DefWindowProcW(handle, msg, wparam, lparam);
		}

		LRESULT window_proc(UINT msg, WPARAM wparam, LPARAM lparam)
		{
			//Instance window procedure; most messages should be handled here and derrived classes can override it.
			switch (msg)
			{
			case WM_CLOSE:
				DestroyWindow(m_handle);
				return 0;
			case WM_DESTROY:
				PostQuitMessage(0);
				return 0;
			}
			return DefWindowProcW(m_handle, msg, wparam, lparam);
		}

		/***** Other members *****/
		window &operator =(window const &) = delete;
		[[nodiscard]] operator HWND() const noexcept { return m_handle; }
		inline ~window() noexcept { if (m_handle) DestroyWindow(m_handle); }
	};
}