module;

#include <windows.h>
#include <cassert>

export module pige.winapi:exceptions;

import std;

export namespace pige::winapi
{
	//Helper class for handling errors, code given can mostly be used as-is
	class winapi_error
	{
		DWORD m_error_code;
		std::stacktrace m_trace;

	public:
		inline explicit winapi_error(
				DWORD error_code = GetLastError() /* error codes for most WinAPI functions are retrived by calling
													 GetLastError, see:
								https://learn.microsoft.com/en-us/windows/win32/Debug/retrieving-the-last-error-code and
				   https://learn.microsoft.com/en-us/windows/win32/api/errhandlingapi/nf-errhandlingapi-getlasterror */,

				std::stacktrace trace = std::stacktrace::current() /* C++23, for C++20 std::source_location could be
																	  used instead*/) noexcept
			: m_error_code{ error_code }, m_trace{ std::move(trace) }
		{ }

		[[nodiscard]] std::wstring what() const noexcept
		{
			std::wstring message;
			LPWSTR buffer = nullptr;
			//Various string allocations in the code bellow technically can throw, but we are already handling
			//an exception, so it's best to catch them. Not much can be done besides ignoring them.
			try
			{
				//Format stack trace as string
				auto trace_str = to_string(m_trace);

				//The stack trace is formatted as a narrow string, but the error description we'll obtain later is
				//a wide string, so we must perform conversion. C++ currently doesn't have a good way to do it, so
				//instead we have to rely on WinAPI function MutliByteToWideChar, see:
				//https://learn.microsoft.com/en-us/windows/win32/api/stringapiset/nf-stringapiset-multibytetowidechar

				std::wstring trace_wstr;

				//Sanity check in case the length of the stack trace exceeds the range of int
				assert(trace_str.size() <= std::numeric_limits<int>::max());

				//obtain the length of wide string after conversion
				auto converted_length = MultiByteToWideChar(CP_UTF8, MB_PRECOMPOSED, trace_str.data(),
					static_cast<int>(trace_str.size()), nullptr, 0);

				//pre-allocate space for converted text
				trace_wstr.resize(converted_length);

				//store converted wide string
				MultiByteToWideChar(CP_UTF8, MB_PRECOMPOSED, trace_str.data(), static_cast<int>(trace_str.size()),
					trace_wstr.data(), converted_length);

				//MultiByteToWideChar might fail, but we can't do much in that case. We are already handling 
				//an exception, so we will continue with emtpy stack trace string

				//Descriptions of WinAPI error codes can be obtained from FormatMessageW using
				//FORMAT_MESSAGE_FROM_SYSTEM flag
				if (FormatMessageW(
					FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
					nullptr, m_error_code, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
					reinterpret_cast<LPWSTR>(&buffer), 0, nullptr))
				{
					//format the message
					message = std::format(L"Error: {}in:\n{}", buffer, trace_wstr);
				}
				else
				{
					message = std::format(L"Error code {} in\n{}", m_error_code, trace_wstr);
				}
			}
			catch (...) {}
			LocalFree(buffer);
			return message;
		}
	};
}