module;

#include <windows.h>

export module pige.winapi:window;

import std;
export import :exceptions;

export namespace pige::winapi {

	class window
	{
	private:

		static void register_class(HINSTANCE instance, LPCWSTR class_name)
		{
			WNDCLASSEXW wcx{
				.cbSize = sizeof(WNDCLASSEXW),
				.style = CS_VREDRAW | CS_HREDRAW,
				.lpfnWndProc = window_proc,
				.hInstance = instance,
				.hCursor = LoadCursorW(nullptr, IDC_ARROW),
				.hbrBackground = static_cast<HBRUSH>(GetStockObject(WHITE_BRUSH)),

				//we can comment out the following line to test our error handling,
				//since RegisterClassExW fails if no class name is given
				.lpszClassName = class_name,
			};
			if (!RegisterClassExW(&wcx))
				throw winapi_error{};
		}

		[[nodiscard]] static bool is_class_registered(HINSTANCE instance, LPCWSTR class_name)
		{
			WNDCLASSEXW wcx{};
			return GetClassInfoExW(instance, class_name, &wcx);
		}

		HWND m_handle;
	public:
		/***** Constructors *****/
		inline window() : m_handle{ nullptr } { }
		window(window const &) = delete;
		window(HINSTANCE instance, LPCWSTR title)
			: window{ }
		{

			LPCWSTR class_name = L"My Window Class";
			if (!is_class_registered(instance, class_name))
				register_class(instance, class_name);
			if (!CreateWindowExW(0, class_name, title, WS_OVERLAPPEDWINDOW | WS_VISIBLE,
								 CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, nullptr, nullptr, nullptr,
								 reinterpret_cast<LPVOID>(this)))
				throw winapi_error{};
		}

		/***** Window Procedure *****/
		static LRESULT CALLBACK window_proc(HWND handle, UINT msg, WPARAM wparam, LPARAM lparam)
		{
			//exceptions thrown out of window procedure tend to get eaten, we must handle them here, just like in main
			try
			{
				window *w = nullptr;
				if (msg == WM_NCCREATE)
				{
					auto pcs = reinterpret_cast<LPCREATESTRUCTW>(lparam);
					w = reinterpret_cast<window *>(pcs->lpCreateParams);
					w->m_handle = handle;

					//SetWindowLongPtrW returns 0 both when previous value in GWLP_USERDATA was 0 and if it fails
					//The only way to check if error occured is by calling GetLastError().
					//However, SetWindowLongPtrW will not clear that might have occured prior to the call.
					//To avoid accidentally and confusingly indicating that this function failed, when in fact
					//something went wrong before, we must clear the error code manually.
					SetLastError(0);
					if (SetWindowLongPtrW(handle, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(w)) == 0)
					{
						auto error_code = GetLastError();
						//since we have to include try/catch in this function any way, we might as well throw,
						//instead of printing an error message directly.
						if (error_code != 0)
							throw winapi::winapi_error{ error_code };
					}
				}
				else
					w = reinterpret_cast<window *>(GetWindowLongPtrW(handle, GWLP_USERDATA));

				if (w)
				{
					auto r = w->window_proc(msg, wparam, lparam);
					if (msg == WM_NCDESTROY)
						w->m_handle = nullptr;
					return r;
				}
				return DefWindowProcW(handle, msg, wparam, lparam);
			}
			catch (winapi::winapi_error &e)
			{
				//handle our own errors
				OutputDebugStringW(e.what().c_str());
			}
			catch (std::exception &e)
			{
				//handle any standard library errors
				OutputDebugStringA(e.what());
			}
			catch (...)
			{
				//catch-all if some other exception happens
				OutputDebugStringW(L"Unknown Error!");
			}
			PostQuitMessage(0);
			return 0;
		}

		LRESULT window_proc(UINT msg, WPARAM wparam, LPARAM lparam)
		{
			//Add your message handling here
			switch (msg)
			{
			case WM_CLOSE:
				DestroyWindow(m_handle);
				return 0;
			case WM_DESTROY:
				PostQuitMessage(0);
				return 0;
			}
			return DefWindowProcW(m_handle, msg, wparam, lparam);
		}

		/***** Other members *****/
		window &operator =(window const &) = delete;
		[[nodiscard]] operator HWND() const noexcept { return m_handle; }
		inline ~window() noexcept { if (m_handle) DestroyWindow(m_handle); }
	};
}