module;

#include <windows.h>
#include <cassert>

export module pige.winapi:window;

import std;
export import :exceptions;

export namespace pige::winapi {

	class application;

	//Helper struct for defining window class description with some reasonable defaults
	struct class_desc : WNDCLASSEXW
	{
		inline constexpr class_desc(
				HINSTANCE instance,
				LPCWSTR name,
				UINT style = CS_VREDRAW | CS_HREDRAW,
				HBRUSH background = reinterpret_cast<HBRUSH>(COLOR_WINDOW + 1),
				HCURSOR cursor = LoadCursorW(nullptr, IDC_ARROW),
				HICON big_icon = nullptr,
				HICON small_icon = nullptr,
				LPCWSTR menu_name = nullptr,
				int extra_class_data = 0,
				int extra_window_data = 0)
			: WNDCLASSEXW{ 
				.cbSize = sizeof(WNDCLASSEXW),
				.style = style,
				.cbClsExtra = extra_class_data,
				.cbWndExtra = extra_window_data,
				.hInstance = instance,
				.hIcon = big_icon,
				.hCursor = cursor,
				.hbrBackground = background,
				.lpszMenuName = menu_name,
				.lpszClassName = name,
				.hIconSm = small_icon }
		{ }
	};

	//Helper class for specifying CreateWindowExW parameters
	struct window_desc : CREATESTRUCTW
	{
		inline explicit constexpr window_desc(
				LPCWSTR title,
				SIZE size = { CW_USEDEFAULT, CW_USEDEFAULT },
				POINT position = { CW_USEDEFAULT, CW_USEDEFAULT },
				LONG style = WS_OVERLAPPEDWINDOW,
				DWORD extended_style = 0,
				HWND parent = nullptr,
				HMENU menu = nullptr)
			: CREATESTRUCTW{ 
				.hMenu = menu,
				.hwndParent = parent,
				.cy = size.cy,
				.cx = size.cx,
				.y = position.y,
				.x = position.x,
				.style = style,
				.lpszName = title,
				.dwExStyle = extended_style }
		{ }
	};

	//From now on this class can stay as-is. Any modifications to the window's look and behaviour can be done
	//is a sub-class
	class window
	{
	private:
		static void register_class(class_desc &desc)
		{
			//Now we get the whole class description via one parameter. We are going to ignore value of lpfnWndProc
			//field and make sure our static window procedure is used by all windows
			desc.lpfnWndProc = window_proc;
			if (!RegisterClassExW(&desc))
				throw winapi_error{};
		}

		[[nodiscard]] static bool is_class_registered(class_desc const &desc)
		{
			WNDCLASSEXW wcx{};
			if (GetClassInfoExW(desc.hInstance, desc.lpszClassName, &wcx))
			{
				assert(wcx.lpfnWndProc == static_cast<WNDPROC>(window_proc));
				return true;
			}
			return false;
		}

		HWND m_handle;

		//We want the application class to designate which window is the main one, so that closing it closes the
		//whole program. This will be helpful in scenarios with mutliple windows.
		friend class application;
		bool m_is_main = false;

	public:

		static constexpr std::wstring_view default_class_name{ L"My Window Class" };
		inline static class_desc default_class(HINSTANCE instance)
		{
			return class_desc{ instance, default_class_name.data() };
		}

		/***** Constructors *****/
		inline window() : m_handle{ nullptr } { }
		window(window const &) = delete;

		//Allow the window to be move-constructed and move-assigned
		inline window(window &&other) : m_handle{ nullptr } { *this = std::move(other); }

		inline explicit window(HINSTANCE instance, LPCWSTR title) 
			: window{ default_class(instance), window_desc{ title } }
		{ }

	protected:
		//Helper constructors that allow derived classes to change properties of the window or the window class
		//using the structures provided above
		inline window(HINSTANCE instance, window_desc const &desc)
			: window{ default_class(instance), desc }
		{ }
		window(class_desc cls_desc, window_desc const &desc)
			: window{ }
		{
			if (!is_class_registered(cls_desc))
				register_class(cls_desc);
			auto handle = CreateWindowExW(desc.dwExStyle,
										  cls_desc.lpszClassName,
										  desc.lpszName,
										  desc.style,
										  desc.x, desc.y,
										  desc.cx, desc.cy,
										  desc.hwndParent,
										  desc.hMenu,
										  cls_desc.hInstance,
										  this);
			if (!handle)
				throw winapi_error{};
		}

		/***** Window Procedure *****/
		static LRESULT CALLBACK window_proc(HWND handle, UINT msg, WPARAM wparam, LPARAM lparam)
		{
			try
			{
				window *w = nullptr;
				if (msg == WM_NCCREATE)
				{
					auto pcs = reinterpret_cast<LPCREATESTRUCTW>(lparam);
					w = reinterpret_cast<window *>(pcs->lpCreateParams);
					w->m_handle = handle;
					SetLastError(0);
					if (SetWindowLongPtrW(handle, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(w)) == 0)
					{
						auto error_code = GetLastError();
						if (error_code != 0)
							throw winapi_error{error_code};
					}
				}
				else
					w = reinterpret_cast<window *>(GetWindowLongPtrW(handle, GWLP_USERDATA));

				if (w)
				{
					auto r = w->window_proc(msg, wparam, lparam);
					if (msg == WM_NCDESTROY)
						w->m_handle = nullptr;
					return r;
				}
				return DefWindowProcW(handle, msg, wparam, lparam);
			}
			catch (winapi_error &e) {
				OutputDebugStringW(e.what().c_str());
			}
			catch (std::exception &e) {
				OutputDebugStringA(e.what());
			}
			catch (...) {
				OutputDebugStringW(L"Unknown Error!");
			}
			PostQuitMessage(0);
			return 0;
		}

		//Override this function in a derived class to include additional message handling. Don't forget to forward
		//unhandled messages to the base function.
		virtual LRESULT window_proc(UINT msg, WPARAM wparam, LPARAM lparam)
		{
			//Only for the main window should closing it cause the application to exit.
			if (m_is_main) {
				switch (msg)
				{
				case WM_CLOSE:
					DestroyWindow(m_handle);
					return 0;
				case WM_DESTROY:
					PostQuitMessage(0);
					return 0;
				}
			}
			return DefWindowProcW(m_handle, msg, wparam, lparam);
		}

	public:
		/***** Other members *****/
		window &operator =(window const &) = delete;

		//Allow the window to be move-constructed and move-assigned
		window &operator =(window &&other) noexcept
		{
			std::swap(m_handle, other.m_handle);
			std::swap(m_is_main, other.m_is_main);
			//Besides swapping member fields, we need to ensure to update GWLP_USERDATA of the windows
			if (m_handle)
				SetWindowLongPtrW(m_handle, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(this));
			if (other.m_handle)
				SetWindowLongPtrW(other.m_handle, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(&other));
			return *this;
		}
		[[nodiscard]] operator HWND() const noexcept { return m_handle; }
		//In a virtual class, non-trivial destructor should be virtual
		inline virtual ~window() noexcept { if (m_handle) DestroyWindow(m_handle); }
	};
}