module;

#include <windows.h>
#include <dwmapi.h>
#include "resource.h"

export module pige.sample;

export import pige.winapi;
import std;

namespace pige::samples
{
	struct field
	{
		RECT position;
	};

	class board
	{
	public:
		static constexpr LONG columns = 4, rows = 4, margin = 10, field_size = 60;
		static constexpr LONG field_count = rows * columns;
		static constexpr LONG field_stride = field_size + margin;
		static constexpr LONG width = columns * field_stride + margin;
		static constexpr LONG height = rows * field_stride + margin;
		using field_array = std::array<field, field_count>;

		board()
			: m_fields{}
		{
			for (LONG row = 0; row < rows; ++row)
				for (LONG column = 0; column < columns; ++column)
					m_fields[row * columns + column] = field{ {
							.left = column * field_stride + margin,
							.top = row * field_stride + margin,
							.right = (column + 1) * field_stride,
							.bottom = (row + 1) * field_stride
						} };
		}

		field_array const & fields() const { return m_fields; }

	private:
		field_array m_fields;
	};

	class window_2048 : public pige::winapi::window
	{
	private:
		static auto get_desc(LONG style, DWORD extended_style, HWND parent = nullptr)
		{
			RECT size{ 0, 0, board::width, board::height };
			AdjustWindowRectEx(&size, style, false, extended_style);
			return pige::winapi::window_desc{
				L"2048", { size.right - size.left, size.bottom - size.top }, { CW_USEDEFAULT, CW_USEDEFAULT },
				style, extended_style, parent
			};
		}

		static auto &get_class_desc(HINSTANCE instance)
		{
			static pige::winapi::class_desc desc{ 
				nullptr, L"2048 Window", 0, CreateSolidBrush(RGB(250, 247, 238))
			};
			if (!desc.hIcon)
			{
				desc.hIcon = static_cast<HICON>(LoadImageW(instance, MAKEINTRESOURCE(ID_APPICON), IMAGE_ICON,
														   0, 0, LR_SHARED | LR_DEFAULTSIZE));
			}
			desc.hInstance = instance;
			return desc;
		}

	protected:
		window_2048(HINSTANCE instance, LONG style, board *b, DWORD extended_style = 0, HWND parent = nullptr)
			: window{ get_class_desc(instance), get_desc(style, extended_style, parent) },
			  m_field_brush{ CreateSolidBrush(RGB(204, 192, 174)) }, m_board{ b }
		{
			//Right now the board is only used in the constructor, however, rememebering it will be useful
			//in the extended version of the task.
			for (auto &f : m_board->fields())
				CreateWindowExW(0, L"STATIC", nullptr, WS_CHILD | WS_VISIBLE | SS_CENTER, f.position.left,
					f.position.top, f.position.right - f.position.left, f.position.bottom - f.position.top,
					*this, nullptr, instance, nullptr);
		}

		window_2048(window_2048 &&other)
			: window{ std::move(other) }, m_field_brush{ nullptr }, m_board{ nullptr }
		{
			std::swap(m_field_brush, other.m_field_brush);
			std::swap(m_board, other.m_board);
		}

		HWND get_other_window()
		{
			//This could be a pure virtual method since both subclasses would be able to more easily determine
			//what the other window's handle is. But this is almost as simple, so it hardly seems worth it.
			HWND other = GetWindow(*this, GW_OWNER);
			if (!other)
				other = GetWindow(*this, GW_ENABLEDPOPUP);
			return other;
		}

		LRESULT window_proc(UINT msg, WPARAM wparam, LPARAM lparam) override
		{
			switch (msg)
			{
			case WM_CTLCOLORSTATIC:
				return reinterpret_cast<LRESULT>(m_field_brush);
			case WM_WINDOWPOSCHANGED:
				on_window_move(*reinterpret_cast<LPWINDOWPOS>(lparam));
				return 0;
			default:
				return pige::winapi::window::window_proc(msg, wparam, lparam);
			}
		}

	public:
		window_2048 &operator =(window_2048 &&other)
		{
			static_cast<pige::winapi::window&>(*this) = std::move(other);
			std::swap(m_field_brush, other.m_field_brush);
			std::swap(m_board, other.m_board);
		}

		~window_2048() override
		{
			if (m_field_brush)
				DeleteObject(m_field_brush);
		}

	private:

		void on_window_move(WINDOWPOS position)
		{
			static POINT screen_size{ GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN) };
			POINT centre{ board::width / 2, board::height / 2 };
			POINT other_centre = centre;
			ClientToScreen(*this, &centre);
			HWND other = get_other_window();
			ClientToScreen(other, &other_centre);
			POINT new_centre{ screen_size.x - centre.x, screen_size.y - centre.y };
			RECT other_rc;
			GetWindowRect(other, &other_rc);
			POINT other_npos{
				other_rc.left + new_centre.x - other_centre.x,
				other_rc.top + new_centre.y - other_centre.y
			};
			if (other_npos.x != other_rc.left || other_npos.y != other_rc.top)
				SetWindowPos(other, nullptr, other_npos.x, other_npos.y, 0, 0, SWP_NOSIZE | SWP_NOACTIVATE | SWP_NOZORDER);
		}

		HBRUSH m_field_brush;
		board *m_board;
	};

	class main_2048;

	class popup_2048 : public window_2048
	{
	public:
		popup_2048(HINSTANCE instance, HWND parent, board *b)
			: window_2048{ instance, WS_OVERLAPPED | WS_CAPTION | WS_VISIBLE, b, WS_EX_LAYERED, parent }
		{
			SetLayeredWindowAttributes(*this, 0, 255, LWA_ALPHA);
		}

		LRESULT window_proc(UINT msg, WPARAM wparam, LPARAM lparam) override
		{
			auto result = window_2048::window_proc(msg, wparam, lparam);
			if (msg == WM_WINDOWPOSCHANGED)
			{
				HWND main = get_other_window();
				RECT main_rc, popup_rc, intersection;
				DwmGetWindowAttribute(*this, DWMWA_EXTENDED_FRAME_BOUNDS, &popup_rc, sizeof(RECT));
				DwmGetWindowAttribute(main, DWMWA_EXTENDED_FRAME_BOUNDS, &main_rc, sizeof(RECT));
				IntersectRect(&intersection, &main_rc, &popup_rc);
				BYTE alpha = IsRectEmpty(&intersection) ? 255 : 128;
				SetLayeredWindowAttributes(*this, 0, alpha, LWA_ALPHA);
			}
			return result;
		}
	};

	class main_2048 : public window_2048
	{
		popup_2048 m_popup;

	public:

		main_2048(HINSTANCE instance, board *b)
			: window_2048{ instance, WS_OVERLAPPEDWINDOW ^ WS_SIZEBOX ^ WS_MAXIMIZEBOX, b },
			  m_popup { instance, *this, b }
		{ }
	};

	export class app2048 : protected pige::winapi::application
	{
	private:
		board m_board;
		main_2048 m_main_window;

	public:
		app2048(HINSTANCE instance)
			: application{ instance }, m_board{}, m_main_window{ instance, &m_board }
		{ }

		using application::instance;

		int run(int show_command)
		{
			ShowWindow(m_main_window, show_command);
			return application::run(m_main_window);
		}
	};
}