import pige.winapi;
import std;

#include <windows.h>

//Base class for the three kinds of windows in this example
//Hitting appropriate keys on the keyboard allows the window and class styles to be modified on the fly.
//For descriptions of various styles used here, see:
//https://learn.microsoft.com/en-us/windows/win32/winmsg/window-styles ,
//https://learn.microsoft.com/en-us/windows/win32/winmsg/extended-window-styles and
//https://learn.microsoft.com/en-us/windows/win32/winmsg/window-class-styles

class style_window : public pige::winapi::window
{
	UINT  cls_style = 0;
	LONG  wnd_style = 0;
	DWORD wex_style = 0;

	void on_char(wchar_t ch)
	{
		//Extended window styles
		if (iswupper(ch))
		{
			switch (ch)
			{
			default:
				return;
			case L'D':
				wex_style ^= WS_EX_DLGMODALFRAME;
				break;
			case L'T':
				wex_style ^= WS_EX_TOOLWINDOW;
				break;
			case L'W':
				//Mentioned in the documentation, but doesn't have any visual effect nowadays
				wex_style ^= WS_EX_WINDOWEDGE;
				break;
			case L'C':
				wex_style ^= WS_EX_CLIENTEDGE;
				break;
			case L'H':
				wex_style ^= WS_EX_CONTEXTHELP;
				break;
			case L'S':
				wex_style ^= WS_EX_STATICEDGE;
				break;
			case L'A':
				wex_style ^= WS_EX_APPWINDOW;
				break;

				//trying to modify WS_EX_TOPMOST style in the same manner will not work. To make window top-most, call:
				//SetWindowPos(*this, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE)
				//or to make it non-topmost:
				//SetWindowPos(*this, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE)
			}
			SetWindowLongPtrW(*this, GWL_EXSTYLE, wex_style);
		}
		//Window styles
		else if (iswlower(ch))
		{
			switch (ch)
			{
			default:
				return;
			case L'b':
				wnd_style ^= WS_BORDER;
				break;
			case L'd':
				wnd_style ^= WS_DLGFRAME;
				break;
			case L's':
				wnd_style ^= WS_SYSMENU;
				break;
			case L't':
				wnd_style ^= WS_THICKFRAME;
				break;
			case L'i':
				wnd_style ^= WS_MINIMIZEBOX;
				break;
			case L'm':
				wnd_style ^= WS_MAXIMIZEBOX;
				break;
			}
			SetWindowLongPtrW(*this, GWL_STYLE, wnd_style);
		}
		//Class styles
		else if (iswdigit(ch))
		{
			switch (ch)
			{
			default:
				return;
			case L'1':
				cls_style ^= CS_NOCLOSE;
				break;
				//no easy way to adjust CS_DROPSHADOW on the fly - change it in the get_class functions
				//of the sub-classes. Remember, it has no effect for windows with WS_CHILD.
			}
			SetClassLongPtrW(*this, GCL_STYLE, cls_style);
		}
		else return;
		SetWindowPos(*this, nullptr, 0, 0, 0, 0,
					 SWP_FRAMECHANGED | SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE | SWP_NOZORDER);
	}

public:
	style_window(pige::winapi::class_desc const &cls_desc, pige::winapi::window_desc const &wnd_desc)
		: window{ cls_desc, wnd_desc }, cls_style{ cls_desc.style },
		  wnd_style{ wnd_desc.style }, wex_style{ wnd_desc.dwExStyle }
	{ }

protected:
	LRESULT window_proc(UINT msg, WPARAM wparam, LPARAM lparam) override
	{
		switch (msg)
		{
		case WM_CHAR:
			on_char(static_cast<wchar_t>(wparam));
			return 0;

			//by default only the parent window recieves focus, we'll force it tho whichever window is clicked
			//so that we can also mess with the child window.
		case WM_MOUSEACTIVATE:
			SetFocus(*this);
			return MA_ACTIVATE;
		case WM_LBUTTONDOWN:
		case WM_RBUTTONDOWN:
		case WM_MBUTTONDOWN:
			SetFocus(*this);
			return 0;
		default:
			return window::window_proc(msg, wparam, lparam);
		}
	}
};

//Popup windows are intended to be used as auxiliary, often short-lived, windows (e.g. menu fly-outs, message boxes,
//etc.). However, there are not that many differences between what can be achieved with a popup window and a regular
//overlapped window. They have different defaults, but creation parameters can be adjusted to get moslty the same
//results for both. Most of the differences in behaviour of this window have more to do with the fact, that it is an
//owned window, and not because it's a popup.
class popup_styles : public style_window
{
	static auto get_class(HINSTANCE instance)
	{
		return pige::winapi::class_desc{ instance, L"Styled Popup"/*, CS_DROPSHADOW */ };
	}

	static auto get_desc(HWND parent)
	{
		LONG style = WS_POPUP | WS_CAPTION | WS_VISIBLE;
		return pige::winapi::window_desc{
			L"Popup Style", { 300, 200 }, { CW_USEDEFAULT, CW_USEDEFAULT },
			style, WS_EX_APPWINDOW, parent
		};
	}

public:
	popup_styles(HINSTANCE instance, HWND parent)
		: style_window{ get_class(instance), get_desc(parent) }
	{ }
};

//Child window is a window contained within its parent. They are intended to be used as, often interactive,
//ui elements placed on the parent (e.g. controls like buttons etc.), but they can be forced to behave almost like
//top-level windows.
class child_styles : public style_window
{
	static auto get_class(HINSTANCE instance)
	{
		return pige::winapi::class_desc{ instance, L"Styled Child", 0, reinterpret_cast<HBRUSH>(COLOR_BTNFACE + 1) };
	}

	static auto get_desc(HWND parent)
	{
		return pige::winapi::window_desc{ 
			L"Child Style", { 300, 200 }, { 200, 100 }, WS_CHILD | WS_CAPTION | WS_VISIBLE, 0, parent
		};
	}

public:
	child_styles(HINSTANCE instance, HWND parent)
		: style_window{ get_class(instance), get_desc(parent) }
	{ }
};


class main_styles : public style_window
{
	static auto get_class(HINSTANCE instance)
	{
		return pige::winapi::class_desc{ instance, L"Styled Window" /*, CS_DROPSHADOW */ };
	}

	static auto get_desc()
	{
		return pige::winapi::window_desc{ L"Overlapped Style", { CW_USEDEFAULT, CW_USEDEFAULT }, { CW_USEDEFAULT, CW_USEDEFAULT }, WS_OVERLAPPEDWINDOW | WS_VISIBLE };
	}

	child_styles m_child;
	popup_styles m_popup;

public:
	explicit main_styles(HINSTANCE instance)
		: style_window{ get_class(instance), get_desc() }, m_child{ instance, *this }, m_popup{ instance, *this }
	{ }
};

int WINAPI wWinMain(HINSTANCE instance, HINSTANCE, LPWSTR, int show_command)
{
	try {
		main_styles m_window{ instance };
		ShowWindow(m_window, show_command);
		pige::winapi::application app{ instance };
		return app.run(m_window);
	}
	catch (pige::winapi::winapi_error &e)
	{
		//handle our own errors
		OutputDebugStringW(e.what().c_str());
	}
	catch (std::exception &e)
	{
		//handle any standard library errors
		OutputDebugStringA(e.what());
	}
	catch (...)
	{
		//catch-all if some other exception happens
		OutputDebugStringW(L"Unknown Error!");
	}
	return 1;
}